//----------------------------------------------------------------------------------
// File:        tegra_khr_apps_sdk\samples\opengles20\src\win_egl_simple\win_egl_simple.cpp
// SDK Version: 1.0.0 
// Email:       tegradev@nvidia.com
// Forum:       http://developer.nvidia.com/tegra/forum
//
// Copyright (c) 2007-2010 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

/*
 * win_egl_simple.cpp
 *
 * Just one triangle, drawn using only WinCE, EGL, and GL ES 2.0
 * 
 */

#define MULTIDISPLAY 1

#include <Windows.h>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdio.h>
#include "GLES2/gl2.h"
#include "GLES2/gl2ext.h"
#include "EGL/egl.h"
#include "EGL/eglext.h"
#include "EGL/eglplatform.h"
#if MULTIDISPLAY
#include "multidisplay.h"
#endif

// WinCE SDK doesn't seem to declare PI
const float NV_PI = 3.14159265358979323846f;

// We should use VBOs, but for this tiny example, we go for simplicity over
// efficiency.  But all real rendering apps should use VBOs
#define ROOT_3_OVER_2 0.8660254f
#define ROOT_3_OVER_6 (ROOT_3_OVER_2/3.0f)
static GLfloat s_vert[6] = { 0.5f, -ROOT_3_OVER_6,
                          -0.5f, -ROOT_3_OVER_6,
                           0.0f,  ROOT_3_OVER_2 - ROOT_3_OVER_6 };
static GLfloat s_col [12] = { 1.0, 0.0, 0.0, 1.0,
                0.0, 1.0, 0.0, 1.0,
                0.0, 0.0, 1.0, 1.0};

// Cached handles to the shader program and to the rotation uniform
// inside of the shader program
static GLuint s_programObj = 0;
static GLuint s_rotUniformLocation = 0;
static float s_angle = 0.0f;
static bool s_spin = true;

EGLDisplay s_eglDisplay = EGL_NO_DISPLAY;
EGLSurface s_eglSurface = EGL_NO_SURFACE;
EGLContext s_eglContext = EGL_NO_CONTEXT;

const static char s_fragShaderSource[] =
"varying vec4 col_var;\n"
"void main()\n"
"{\n"
"    gl_FragColor = col_var;\n"
"}";

const static char s_vertShaderSource[] =
"uniform vec2 rot; // rot.xy = {cos(theta), sin(theta)}\n"
"attribute vec2 pos_attr;\n"
"attribute vec4 col_attr;\n"
"varying vec4 col_var;\n"
"void main()\n"
"{\n"
"    gl_Position.x = rot.x * pos_attr.x - rot.y * pos_attr.y;\n"
"    gl_Position.y = rot.y * pos_attr.x + rot.x * pos_attr.y;\n"
"    gl_Position.z = 0.0;\n"
"    col_var = col_attr;\n"
"}";

bool LoadSourceShader(GLuint shader, const char* source)
{
    int len = strlen(source);
    glShaderSource(shader, 1, &source, &len);

    glCompileShader(shader);
    GLint success = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
    if(!success) 
    {
        printf("\n>>> Couldn't compile shader! Exiting...\n\n");
        return false;
    }

    return true;
}

static bool InitEGL(EGLNativeWindowType nativeWindow)
{
    EGLint contextAttrs[] = 
    {
        EGL_CONTEXT_CLIENT_VERSION, 2,
        EGL_NONE
    };

    s_eglDisplay = eglGetDisplay(EGL_DEFAULT_DISPLAY);
    if (s_eglDisplay == EGL_NO_DISPLAY) 
    {
        printf("\n>>> EGL failed to obtain display!  Exiting...\n\n");
        return false;
    }

    if (!eglInitialize(s_eglDisplay, 0, 0)) 
    {
        printf("\n>>> EGL failed to initialize!  Exiting...\n\n");
        return false;
    }

    EGLBoolean exactMatchOnly	= 0;
    EGLBoolean hasDepth			= 1;
    EGLBoolean hasStencil		= 0;

    typedef struct
    {
        EGLint redSize;
        EGLint greenSize;
        EGLint blueSize;
        EGLint alphaSize;
        EGLint depthSize;
        EGLint stencilSize;
    } NvTglConfigRequest;
    NvTglConfigRequest request;

    memset(&request, 0, sizeof(request));

    request.redSize = 5;
    request.greenSize = 6;
    request.blueSize = 5;
    request.alphaSize = 0;

    if (hasDepth)
        request.depthSize = 1;   // accept any size

    if (hasStencil)
        request.stencilSize = 1;   // accept any size

    #define MAX_CONFIG_ATTRS_SIZE 32
    #define MAX_MATCHING_CONFIGS 64
    EGLint configAttrs[MAX_CONFIG_ATTRS_SIZE];
    EGLint i = 0;
    // construct attribute request
    configAttrs[i] = EGL_RENDERABLE_TYPE;
    configAttrs[++i] = EGL_OPENGL_ES2_BIT;
    configAttrs[++i] = EGL_BUFFER_SIZE;
    configAttrs[++i] = request.redSize + request.greenSize + 
                       request.blueSize + request.alphaSize;
    configAttrs[++i] = EGL_RED_SIZE;
    configAttrs[++i] = request.redSize;
    configAttrs[++i] = EGL_GREEN_SIZE;
    configAttrs[++i] = request.greenSize;
    configAttrs[++i] = EGL_BLUE_SIZE;
    configAttrs[++i] = request.blueSize;
    configAttrs[++i] = EGL_ALPHA_SIZE;
    configAttrs[++i] = request.alphaSize;
    configAttrs[++i] = EGL_DEPTH_SIZE;
    configAttrs[++i] = request.depthSize;
    configAttrs[++i] = EGL_STENCIL_SIZE;
    configAttrs[++i] = request.stencilSize;
    configAttrs[++i] = EGL_NONE;

    assert(i < MAX_CONFIG_ATTRS_SIZE);

    // choose configs
    EGLConfig matchingConfigs[MAX_MATCHING_CONFIGS];
    EGLint numMatchingConfigs = 0;
    EGLBoolean success = eglChooseConfig(s_eglDisplay, configAttrs, matchingConfigs, 
        MAX_MATCHING_CONFIGS, &numMatchingConfigs);
    if (!success)
    {
        printf("eglChooseConfig failed!\n");
        return false;
    }
    
    if (numMatchingConfigs == 0)
    {
        printf("No matching EGL configs found!\n");
        return false;
    }

    EGLConfig config;

    // look for exact match
    for (i = 0; i < numMatchingConfigs; i++)
    {
        EGLConfig testConfig = matchingConfigs[i];

        // query attributes from config
        NvTglConfigRequest cfg;
        EGLBoolean ok = EGL_TRUE;
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_RED_SIZE, &cfg.redSize);
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_GREEN_SIZE, &cfg.greenSize);
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_BLUE_SIZE, &cfg.blueSize);
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_ALPHA_SIZE, &cfg.alphaSize);
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_DEPTH_SIZE, &cfg.depthSize);
        ok &= eglGetConfigAttrib(s_eglDisplay, testConfig, EGL_STENCIL_SIZE, &cfg.stencilSize);
        if (!ok)
        {
            printf("eglGetConfigAttrib failed!\n");
            break;
        }

        // depthSize == 1 indicates to accept any non-zero depth
        EGLBoolean depthMatches = ((request.depthSize == 1) && (cfg.depthSize > 0)) ||
                       (cfg.depthSize == request.depthSize);

        // stencilSize == 1 indicates to accept any non-zero stencil
        EGLBoolean stencilMatches = ((request.stencilSize == 1) && (cfg.stencilSize > 0)) ||
                         (cfg.stencilSize == request.stencilSize);
        
        // compare to request
        if (cfg.redSize == request.redSize &&
                cfg.greenSize == request.greenSize &&
                cfg.blueSize == request.blueSize &&
                cfg.alphaSize == request.alphaSize &&
                depthMatches && stencilMatches)
        {
            config = testConfig;
            break;
        }
    }
    
    if (i >= numMatchingConfigs)
    {
        // no exact match was found
        if (exactMatchOnly)
        {
            printf("No exactly matching EGL configs found!\n");
            return false;
        }

        // just return any matching config (ie, the first one)
        config = matchingConfigs[0];
    }

    s_eglContext = eglCreateContext(s_eglDisplay, config, EGL_NO_CONTEXT, contextAttrs);
    if (!s_eglContext) 
    {
        printf("\n>>> EGL failed to create context!  Exiting...\n\n");
        return false;
    }

    s_eglSurface = eglCreateWindowSurface(s_eglDisplay, config, nativeWindow, 0);
    if (!s_eglSurface) 
    {
        printf("\n>>> EGL failed to create window surface!  Exiting..\n\n");
        return false;
    }

    if (!eglMakeCurrent(s_eglDisplay, s_eglSurface, s_eglSurface, s_eglContext))
    {
        printf("\n>>> EGL couldn't make context current!  Exiting...\n\n");
        return false;
    }

    if (!eglSwapInterval(s_eglDisplay, 0)) 
    {
        printf("\n>>> EGL couldn't set swap interval! Exiting...\n\n");
        return false;
    }

    return true;
}

static bool InitGraphicsState(void)
{
    GLuint vsObj, fsObj;

    vsObj = glCreateShader(GL_VERTEX_SHADER);
    fsObj = glCreateShader(GL_FRAGMENT_SHADER);
    s_programObj = glCreateProgram();
    glAttachShader(s_programObj, vsObj);
    glAttachShader(s_programObj, fsObj);

    if (!LoadSourceShader(vsObj, s_vertShaderSource))
    {
        printf("\n>>> Couldn't load source vertex shader from file! Exiting...\n\n");
        return false;
    }
    if (!LoadSourceShader(fsObj, s_fragShaderSource))
    {
        printf("\n>>> Couldn't load source fragment shader from file! Exiting...\n\n");
        return false;
    }

    glBindAttribLocation(s_programObj, 0, "pos_attr");
    glBindAttribLocation(s_programObj, 1, "col_attr");

    glLinkProgram(s_programObj);
    GLint success = 0;
    glGetProgramiv(s_programObj, GL_LINK_STATUS, &success);
    if(!success) 
    {
        printf("\n>>> Couldn't link shader! Exiting...\n\n");
        return false;
    }

    glUseProgram(s_programObj);

    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, s_vert);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, s_col);

    s_rotUniformLocation = glGetUniformLocation(s_programObj, "rot");

    return true;
}


static void Display(void)
{
    // This is not the safest or best way to handle timing, but this code
    // is only added to make the triangle rotate at a basically constant
    // rate, independent of the target (Win32) platform
    static DWORD lastTime = GetTickCount();
    DWORD newTime = GetTickCount();
    DWORD delta = newTime - lastTime;

    // Handle rollover
    if (newTime < lastTime)
        delta = 0;

    lastTime = newTime;

    if (s_spin)
        s_angle += delta * 0.01f * NV_PI / 180.0f;

    glClear(GL_COLOR_BUFFER_BIT);

    // we only need to pass cos and sin to the shader.
    glUniform2f(s_rotUniformLocation, cos(s_angle), sin(s_angle));

    glDrawArrays(GL_TRIANGLES, 0, 3);

    eglSwapBuffers(s_eglDisplay, s_eglSurface);
}


static bool ShutdownEGL()
{
    eglMakeCurrent(s_eglDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);

    if(eglDestroyContext(s_eglDisplay, s_eglContext))
        printf("eglDestroyContext SUCCESS\n");
    if(eglDestroySurface(s_eglDisplay, s_eglSurface))
        printf("eglDestroySurface SUCCESS\n");
    if (eglTerminate(s_eglDisplay))
        printf("eglTerminate SUCCESS\n");

    return true;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{                                                                       
    switch (message)
    {
        case WM_LBUTTONDOWN:  
        case WM_DESTROY:                                                    
            PostQuitMessage(0);                                             
            break;

        default:                                                            
            return DefWindowProc(hWnd, message, wParam, lParam);            
    }                                                                   
    return 0;                                                           
}    


ATOM MyRegisterClass()
{
    WNDCLASS wndclass;

#if _WIN32_WCE
    wndclass.style          = 0;
#else
    wndclass.style          = CS_OWNDC;
#endif
    wndclass.lpfnWndProc    = WndProc;
    wndclass.cbClsExtra     = 0;
    wndclass.cbWndExtra     = 0;
    wndclass.hInstance      = (HINSTANCE)GetModuleHandle(NULL);
    wndclass.hIcon          = NULL;
    wndclass.hCursor        = LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground  = (HBRUSH) GetStockObject (BLACK_BRUSH);
    wndclass.lpszMenuName   = NULL;
    wndclass.lpszClassName  = TEXT("EGLWndClass");
    return RegisterClass(&wndclass);
}


#ifdef _WIN32_WCE
int WINAPI WinMain( HINSTANCE,
                    HINSTANCE,
                    LPTSTR,
                    int)
#else
int main (int, const char * const*)
#endif
{
    int failure = 1;

    MyRegisterClass();

    HWND hwnd = CreateWindow(
        TEXT("EGLWndClass"),
        TEXT("win_egl_simple"),
        WS_DLGFRAME|WS_CAPTION|WS_THICKFRAME|WS_SYSMENU,
        CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
        NULL,
        NULL,
        (HINSTANCE)GetModuleHandle(NULL),
        NULL);

    if (!hwnd)
        return -1;

    //EnableWindow(hwnd, TRUE);

#if 0
    RECT area;
    area.left = 0;
    area.top = 0;
#ifdef _WIN32_WCE
    area.right = GetSystemMetrics(SM_CXSCREEN);
    area.bottom = GetSystemMetrics(SM_CYSCREEN);

    SetWindowLong(hwnd, GWL_STYLE, WS_POPUP);

    SetWindowPos(hwnd, HWND_TOPMOST,
                 area.left, area.top,
                 area.right, area.bottom,
                 SWP_FRAMECHANGED);
#else
    area.right = 800;
    area.bottom = 480;

    SetWindowPos(hwnd, HWND_TOP,
                 area.left, area.top,
                 area.right, area.bottom,
                 SWP_NOMOVE);
#endif
#endif

    /* set as foreground window to give this app focus in case it doesn't have it */
    SetForegroundWindow(hwnd);
    ShowWindow(hwnd, SW_SHOWNORMAL);

#if MULTIDISPLAY
    HANDLE hMD, hOverlay;
    mdDisplayMode old_mode;
    mdOpen(&hMD);
    mdDisplayGetCurrentMode(hMD, MD_DISPLAY_HDMI, &old_mode); // Save current resolution for later use
    mdDisplayActivate(hMD, MD_DISPLAY_HDMI, NULL, 0); // Set HDMI to it's native resolution
    mdOverlayCreate(hMD, MD_DISPLAY_HDMI, &hOverlay); // Create Overlay on HDMI Screen
    EGLNativeWindowType nativeWindow = (EGLNativeWindowType)hOverlay;
#else
    EGLNativeWindowType nativeWindow = (EGLNativeWindowType)hwnd;
#endif

    if (!InitEGL(nativeWindow))
        return failure;

    glClearColor(1.0f, 1.0f, 0.0f, 0.0f);

    if (!InitGraphicsState())
        return failure;
    
    bool quit = false;

    // Main message loop:
    while (!quit)
    {
        Display();

        /* relay message queue messages to windowproc's */
        MSG msg;
        while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&msg);

            if (msg.message == WM_QUIT)
            {
                quit = true;
                break;
            }

            DispatchMessage(&msg);
        }
    }    

    ShutdownEGL();

    if(DestroyWindow (nativeWindow))
        printf("DestroyWindow SUCCESS\n");

#if MULTIDISPLAY
    mdOverlayUnref(hMD, hOverlay); // Ureference, destroy Overlay
    mdDisplayActivate(hMD, MD_DISPLAY_HDMI, &old_mode, 0); // Set back old resolution
    mdClose(hMD);
#endif

    return 0;
}
